import logging
import sys
import traceback
from pathlib import Path
from typing import Optional

import click

from .suggestions import ALL_SUGGESTIONS
from .types import Env

LOG = logging.getLogger(__name__)


@click.command()
@click.option("--verbose", "-v", is_flag=True, help="Show debug output")
@click.option(
    "--autoapply", "-a", is_flag=True, help="Autoapply instead of showing diff"
)
@click.option("--only", help="Just run this one class")
@click.argument("path")
def main(verbose: bool, autoapply: bool, only: Optional[str], path: str) -> None:
    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.ERROR,
        format="%(asctime)-15s %(levelname)-8s %(name)s:%(lineno)s %(message)s",
    )

    env = Env(Path(path))
    rc = 0

    if only:
        selected = [cls for cls in ALL_SUGGESTIONS if cls.__name__ == only]
        if not selected:
            LOG.error(f"Only value {only!r} did not match anything")
            rc |= 2
    else:
        selected = ALL_SUGGESTIONS

    for cls in selected:
        LOG.info(f"Running {cls.__name__}")
        try:
            cls().check(env, autoapply=autoapply)
        except Exception as e:
            rc |= 2
            if verbose:
                traceback.print_exc()
            else:
                print(repr(e))

    sys.exit(rc)


if __name__ == "__main__":  # pragma: no cover
    main()



import io
import logging
from typing import Dict, Optional

import libcst as cst
from moreorless.click import echo_color_unified_diff

import imperfect

from ..metadata import Literal, SetupCallAnalyzer, SetupCallTransformer
from ..setup_and_metadata import SETUP_ARGS
from ..types import BaseSuggestion, Env, SectionWriter

LOG = logging.getLogger(__name__)


class UseDeclarativeConfig(BaseSuggestion):
    def check(self, env: Env, autoapply: bool = False) -> None:
        # TODO: This duplicates some of the logic in
        # opine.metadata.from_setup_py because we don't want a Distribution and
        # want to get the cst object back.

        setup_py = env.base_path / "setup.py"
        if not setup_py.exists():
            return
        module = cst.parse_module(setup_py.read_text())
        analyzer = SetupCallAnalyzer()
        wrapper = cst.MetadataWrapper(module)
        wrapper.visit(analyzer)

        if not analyzer.found_setup:
            self.skip("Could not find setup() call in setup.py")

        setup_cfg = env.base_path / "setup.cfg"
        if setup_cfg.exists():
            cfg_data = setup_cfg.read_text()
        else:
            cfg_data = ""

        cfg = imperfect.parse_string(
            cfg_data + ("\n" if not cfg_data.endswith("\n") else "")
        )
        setup_args_dict = {t.keyword: t for t in SETUP_ARGS}
        keywords_to_change: Dict[str, Optional[cst.CSTNode]] = {}
        for k, v in analyzer.saved_args.items():
            # This is a cursory check on the value; really we should get an
            # AlmostTooComplicated or TooComplicated that works well.
            if (
                k in setup_args_dict
                and isinstance(v, Literal)
                and "??" not in str(v.value)
            ):
                LOG.debug(f"{k}: can move {v.value!r}")
                keywords_to_change[k] = None
                if setup_args_dict[k].cfg.writer_cls is SectionWriter:
                    for k2, v2 in v.value.items():
                        cfg.set_value(
                            setup_args_dict[k].cfg.section,
                            k2,
                            setup_args_dict[k].cfg.writer_cls().to_ini(v2),
                        )
                else:
                    cfg.set_value(
                        setup_args_dict[k].cfg.section,
                        setup_args_dict[k].cfg.key,
                        setup_args_dict[k].cfg.writer_cls().to_ini(v.value),
                    )
            elif k in setup_args_dict:
                LOG.error(f"{k}: too complicated")
            else:
                LOG.debug(f"{k}: unknown keyword")

        buf = io.StringIO()
        cfg.build(buf)

        echo_color_unified_diff(cfg_data, buf.getvalue(), "setup.cfg")

        old_code = module.code
        transformer = SetupCallTransformer(
            cst.ensure_type(analyzer.setup_node, cst.Call), keywords_to_change
        )
        new_module = wrapper.visit(transformer)
        new_code = new_module.code

        echo_color_unified_diff(old_code, new_code, "setup.py")

        # TODO validate that the right args got removed
        # print(f"And remove {sorted(keywords_to_change)}")

        if autoapply:
            setup_cfg.write_text(buf.getvalue())
            setup_py.write_text(new_code)
        elif keywords_to_change:
            print("Rerun with -a instead to apply")



"""
This is mostly compatible with pkginfo's metadata classes.
"""

import json
import logging
import sys
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple

import libcst as cst
import pkginfo.distribution
from libcst.metadata import ParentNodeProvider, QualifiedNameProvider, ScopeProvider

LOG = logging.getLogger(__name__)


# TODO: pkginfo isn't typed
class Distribution(pkginfo.distribution.Distribution):  # type: ignore
    # These are not actually part of the metadata, see PEP 566
    setup_requires: Sequence[str] = ()
    tests_require: Sequence[str] = ()
    extras_require: Dict[str, Sequence[str]] = {}
    use_scm_version: Optional[bool] = None
    zip_safe: Optional[bool] = None
    include_package_data: Optional[bool] = None
    test_suite: str = ""
    namespace_packages: Sequence[str] = ()

    def _getHeaderAttrs(self) -> Sequence[Tuple[str, str, bool]]:
        # Until I invent a metadata version to include this, do so
        # unconditionally.
        return tuple(super()._getHeaderAttrs()) + (
            ("Setup-Requires", "setup_requires", True),
            ("Tests-Require", "tests_require", True),
            ("???", "extras_require", False),
            ("Use-SCM-Version", "use_scm_version", False),
            ("Zip-Safe", "zip_safe", False),
            ("Test-Suite", "test_suite", False),
            ("Include-Package-Data", "include_package_data", False),
            ("Namespace-Package", "namespace_packages", True),
        )


# setup(kwarg=) -> Distribution key
MAPPING = {
    "name": "name",
    "version": "version",
    "author": "author",
    "author_email": "author_email",
    "maintainer": "maintainer",
    "maintainer_email": "maintainer_email",
    "license": "license",
    "description": "summary",
    "long_description": "description",
    "long_description_content_type": "description_content_type",
    "install_requires": "requires_dist",
    "requires": "requires",
    "python_requires": "requires_python",
    "url": "home_page",
    "download_url": "download_url",
    "project_urls": "project_urls",
    "keywords": "keywords",
    "license": "license",
    "platforms": "platforms",
    "use_scm_version": "use_scm_version",
    "setup_requires": "setup_requires",
    "tests_require": "tests_require",
    "extras_require": "extras_require",
    "classifiers": "classifiers",
    "zip_safe": "zip_safe",
    "test_suite": "test_suite",
    "include_package_data": "include_package_data",
    "namespace_packages": "namespace_packages",
}


def from_setup_py(path: Path, markers: Dict[str, Any]) -> Distribution:
    """
    Reads setup.py (and possibly some imports).

    Will not actually "run" the code but will evaluate some conditions based on
    the markers you provide, since much real-world setup.py checks things like
    version, platform, or even `sys.argv` to come up with what it passes to
    `setup()`.

    There should be some other class to read pyproject.toml.

    This needs a path because one day it may need to read other files alongside
    it.
    """

    # TODO: This does not take care of encodings or py2 syntax.
    module = cst.parse_module((path / "setup.py").read_text())

    # TODO: This is not a good example of LibCST integration.  The right way to
    # do this is with a scope provider and transformer, and perhaps multiple
    # passes.

    d = Distribution()
    d.metadata_version = "2.1"

    analyzer = SetupCallAnalyzer()
    wrapper = cst.MetadataWrapper(module)
    wrapper.visit(analyzer)
    if not analyzer.found_setup:
        raise SyntaxError("No simple setup call found")

    for k, v in analyzer.saved_args.items():
        if k in MAPPING:
            if isinstance(v, Literal):
                setattr(d, MAPPING[k], v.value)
            else:
                LOG.warning(f"Want to save {k} but is {type(v)}")
        else:
            LOG.warning(f"Specified {k} but we don't store it")

    return d


@dataclass
class TooComplicated:
    reason: str


@dataclass
class Sometimes:
    # TODO list of 'when' and 'else'
    pass


@dataclass
class Literal:
    value: Any
    cst_node: Optional[cst.CSTNode]


class FileReference:
    def __init__(self, filename: str) -> None:
        self.filename = filename


class SetupCallTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider, QualifiedNameProvider)  # type: ignore

    def __init__(
        self,
        call_node: cst.CSTNode,
        keywords_to_change: Dict[str, Optional[cst.CSTNode]],
    ) -> None:
        self.call_node = call_node
        self.keywords_to_change = keywords_to_change

    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> cst.BaseExpression:
        if original_node == self.call_node:
            new_args = []
            for arg in updated_node.args:
                if isinstance(arg.keyword, cst.Name):
                    if arg.keyword.value in self.keywords_to_change:
                        value = self.keywords_to_change[arg.keyword.value]
                        if value is not None:
                            new_args.append(arg.with_changes(value=value))
                        # else don't append
                    else:
                        new_args.append(arg)
                else:
                    new_args.append(arg)
            return updated_node.with_changes(args=new_args)

        return updated_node


class SetupCallAnalyzer(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider, QualifiedNameProvider)  # type: ignore

    # TODO names resulting from other than 'from setuptools import setup'
    # TODO wrapper funcs that modify args
    # TODO **args
    def __init__(self) -> None:
        super().__init__()
        # TODO Union[TooComplicated, Sometimes, Literal, FileReference]
        self.saved_args: Dict[str, Any] = {}
        self.found_setup = False
        self.setup_node: Optional[cst.CSTNode] = None

    def visit_Call(self, node: cst.Call) -> Optional[bool]:
        names = self.get_metadata(QualifiedNameProvider, node)
        # TODO sometimes there is more than one setup call, we might
        # prioritize/merge...
        if any(
            q.name in ("setuptools.setup", "distutils.core.setup", "setup3lib")
            for q in names
        ):
            self.found_setup = True
            self.setup_node = node
            scope = self.get_metadata(ScopeProvider, node)
            for arg in node.args:
                # TODO **kwargs
                if isinstance(arg.keyword, cst.Name):
                    key = arg.keyword.value
                    value = self.evaluate_in_scope(arg.value, scope)
                    self.saved_args[key] = Literal(value, arg)
                elif arg.star == "**":
                    # kwargs
                    d = self.evaluate_in_scope(arg.value, scope)
                    if isinstance(d, dict):
                        for k, v in d.items():
                            self.saved_args[k] = Literal(v, None)
                    else:
                        # GRR
                        pass
                else:
                    raise ValueError(repr(arg))

            return False

        return None

    BOOL_NAMES = {"True": True, "False": False, "None": None}
    PRETEND_ARGV = ["setup.py", "bdist_wheel"]

    def evaluate_in_scope(self, item: cst.CSTNode, scope: Any) -> Any:
        if isinstance(item, cst.SimpleString):
            return item.evaluated_value
        # TODO int/float/etc
        elif isinstance(item, cst.Name) and item.value in self.BOOL_NAMES:
            return self.BOOL_NAMES[item.value]
        elif isinstance(item, cst.Name):
            name = item.value
            assignments = scope[name]
            for a in assignments:
                # TODO: Only assignments "before" this node matter if in the
                # same scope; really if we had a call graph and walked the other
                # way, we could have a better idea of what has already happened.

                # Assign(
                #   targets=[AssignTarget(target=Name(value="v"))],
                #   value=SimpleString(value="'x'"),
                # )
                # TODO or an import...
                # TODO builtins have BuiltinAssignment
                try:
                    node = a.node
                    if node:
                        parent = self.get_metadata(ParentNodeProvider, node)
                        if parent:
                            gp = self.get_metadata(ParentNodeProvider, parent)
                        else:
                            raise KeyError
                    else:
                        raise KeyError
                except (KeyError, AttributeError):
                    return "??"

                # This presumes a single assignment
                if not isinstance(gp, cst.Assign) or len(gp.targets) != 1:
                    return "??"  # TooComplicated(repr(gp))

                try:
                    scope = self.get_metadata(ScopeProvider, gp)
                except KeyError:
                    # module scope isn't in the dict
                    return "??"

                return self.evaluate_in_scope(gp.value, scope)
        elif isinstance(item, (cst.Tuple, cst.List)):
            lst = []
            for el in item.elements:
                lst.append(
                    self.evaluate_in_scope(
                        el.value, self.get_metadata(ScopeProvider, el)
                    )
                )
            if isinstance(item, cst.Tuple):
                return tuple(lst)
            else:
                return lst
        elif (
            isinstance(item, cst.Call)
            and isinstance(item.func, cst.Name)
            and item.func.value == "dict"
        ):
            d = {}
            for arg in item.args:
                if isinstance(arg.keyword, cst.Name):
                    d[arg.keyword.value] = self.evaluate_in_scope(arg.value, scope)
                # TODO something with **kwargs
            return d
        elif isinstance(item, cst.Dict):
            d = {}
            for el2 in item.elements:
                if isinstance(el2, cst.DictElement):
                    d[self.evaluate_in_scope(el2.key, scope)] = self.evaluate_in_scope(
                        el2.value, scope
                    )
            return d
        elif isinstance(item, cst.Subscript):
            lhs = self.evaluate_in_scope(item.value, scope)
            if isinstance(lhs, str):
                # A "??" entry, propagate
                return "??"

            # TODO: Figure out why this is Sequence
            if isinstance(item.slice[0].slice, cst.Index):
                rhs = self.evaluate_in_scope(item.slice[0].slice.value, scope)
                return lhs.get(rhs, "??")
            else:
                LOG.warning(f"Omit2 {type(item.slice[0].slice)!r}")
                return "??"
        else:
            LOG.warning(f"Omit1 {type(item)!r}")
            return "??"


def main() -> None:
    logging.basicConfig(level=logging.DEBUG)
    for path in sys.argv[1:]:
        try:
            dist = from_setup_py(Path(path), {})
            value = {
                "path": path,
            }

            for k in list(dist):
                if getattr(dist, k):
                    value[k] = getattr(dist, k)

            print(json.dumps(value, indent=2,))
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            print(f"Fail: {path}\n{e!r}", file=sys.stderr)


if __name__ == "__main__":
    main()



import tempfile
import unittest
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from click.testing import CliRunner, Result

from ..main import main

SAMPLE_ENV = {
    "setup.py": "from setuptools import setup\nsetup(name='bar')\n",
    "setup.cfg": "[metadata]\nname=foo\n",
    "tox.ini": "[tox]\nenvlist=py36\n\n[testenv]\ncommands=make\n",
    "coverage.ini": "[run]\na=b\n",
}


class FunctionalTest(unittest.TestCase):
    def _run_and_get_output(
        self, args: List[str], extra_files: Optional[Dict[str, str]] = None
    ) -> Tuple[Result, Dict[str, str]]:
        with tempfile.TemporaryDirectory() as d:
            td_path = Path(d)
            self.files: Dict[str, str] = {}
            self.files.update(SAMPLE_ENV)
            if extra_files:
                self.files.update(extra_files)

            for filename, contents in self.files.items():
                (td_path / filename).write_text(contents)

            runner = CliRunner()
            result = runner.invoke(main, args + [d])

            new_contents: Dict[str, str] = {}
            for p in td_path.glob("*.*"):
                new_contents[p.relative_to(td_path).as_posix()] = p.read_text()

        return result, new_contents

    def test_smoke_no_args(self) -> None:
        result, files = self._run_and_get_output([])
        self.assertEqual(0, result.exit_code)
        self.assertEqual(self.files, files)

    def test_smoke_only_non(self) -> None:
        result, files = self._run_and_get_output(["--only", "Foo"])
        self.assertEqual(2, result.exit_code)
        self.assertEqual(self.files, files)

    def test_smoke_only_one(self) -> None:
        result, files = self._run_and_get_output(["--only", "MoveToxIni"])
        self.assertEqual(0, result.exit_code)
        self.assertEqual(self.files, files)

    def test_smoke_exception(self) -> None:
        result, files = self._run_and_get_output([], {"setup.py": "--"})
        self.assertEqual(2, result.exit_code)
        self.assertEqual(self.files, files)
        self.assertIn('ParserSyntaxError("Incomplete input', result.output)
        self.assertNotIn("Traceback (most recent call last)", result.output)

    def test_smoke_exception_verbose(self) -> None:
        result, files = self._run_and_get_output(["-v"], {"setup.py": "--"})
        self.assertEqual(2, result.exit_code)
        self.assertEqual(self.files, files)
        self.assertIn("ParserSyntaxError: Syntax Error", result.output)
        self.assertIn("Traceback (most recent call last)", result.output)

    def test_smoke_write(self) -> None:
        result, files = self._run_and_get_output(["-a"])
        self.assertEqual(0, result.exit_code)
        self.assertNotIn("tox.ini", files)
        self.assertNotIn("coverage.ini", files)

        self.assertEqual(
            """\
from setuptools import setup
setup()
""",
            files["setup.py"],
        )

        self.assertEqual(
            """\
[metadata]
name=bar

[coverage:run]
a=b

[tox:tox]
envlist=py36

[testenv]
commands=make
""",
            files["setup.cfg"],
        )



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union


# These implement the basic types listed at
# https://setuptools.readthedocs.io/en/latest/setuptools.html#specifying-values
class BaseWriter:
    def to_ini(self, value: Any) -> str:  # pragma: no cover
        raise NotImplementedError


class StrWriter(BaseWriter):
    def to_ini(self, value: str) -> str:
        return value


class ListCommaWriter(BaseWriter):
    def to_ini(self, value: List[str]) -> str:
        if not value:
            return ""
        return "".join(f"\n{k}" for k in value)


class ListCommaWriterCompat(BaseWriter):
    def to_ini(self, value: Union[str, List[str]]) -> str:
        if not value:
            return ""
        if isinstance(value, str):
            value = [value]
        return "".join(f"\n{k}" for k in value)


class ListSemiWriter(BaseWriter):
    def to_ini(self, value: List[str]) -> str:
        if not value:
            return ""
        return "".join(f"\n{k}" for k in value)


# This class is also specialcased
class SectionWriter(BaseWriter):
    def to_ini(self, value: List[str]) -> str:
        if not value:
            return ""
        return "".join(f"\n{k}" for k in value)


class BoolWriter(BaseWriter):
    def to_ini(self, value: bool) -> str:
        return "true" if value else "false"


class DictWriter(BaseWriter):
    def to_ini(self, value: Dict[str, str]) -> str:
        if not value:
            return ""
        return "".join(f"\n{k}={v}" for k, v in value.items())


@dataclass
class SetupCfg:
    section: str
    key: str
    writer_cls: Type[BaseWriter] = StrWriter


@dataclass
class PyProject:
    section: str
    key: str
    # TODO setuptools-only?


@dataclass
class Metadata:
    key: str
    repeated: bool = False


@dataclass
class ConfigField:
    keyword: str
    # TODO type/repeated/etc
    cfg: SetupCfg
    sample_value: Optional[Any] = "foo"


@dataclass
class Env:
    base_path: Path


class BaseSuggestion:
    def check(self, env: Env, autoapply: bool = False) -> None:
        raise NotImplementedError()

    def skip(self, msg: str) -> None:
        raise Exception(msg)  # TODO



import logging

from moreorless.click import echo_color_unified_diff

from imperfect import parse_string

from ..types import BaseSuggestion, Env

LOG = logging.getLogger(__name__)


class MoveToxIni(BaseSuggestion):
    # https://tox.readthedocs.io/en/latest/config.html
    def check(self, env: Env, autoapply: bool = False) -> None:
        tox_ini_path = env.base_path / "tox.ini"
        setup_cfg_path = env.base_path / "setup.cfg"

        if not tox_ini_path.exists():
            return

        tox_ini_text = tox_ini_path.read_text()
        setup_cfg_text = ""
        if setup_cfg_path.exists():
            setup_cfg_text = setup_cfg_path.read_text()

        setup_cfg = parse_string(setup_cfg_text)
        tox_ini = parse_string(tox_ini_text)

        for section in tox_ini.keys():
            # This doesn't iterate and use set_value because that wouldn't
            # preserve comments.
            obj = tox_ini[section]
            if obj.name == "tox":
                obj.name = "tox:tox"

            if obj.name in setup_cfg:
                LOG.error(f"Section {obj.name!r} already exists in setup.cfg")
            else:
                if obj.leading_whitespace == "" and setup_cfg.sections:
                    obj.leading_whitespace = "\n"
                setup_cfg.sections.append(obj)

        new_text = setup_cfg.text
        if new_text != setup_cfg_text:
            echo_color_unified_diff(setup_cfg_text, new_text, "setup.cfg")
            if autoapply:
                setup_cfg_path.write_text(new_text)
                tox_ini_path.unlink()
                print("Written")
            else:
                print("Rerun with -a instead to apply")



import sys
import tempfile
import unittest
from pathlib import Path
from typing import Any, List
from unittest.mock import patch

import libcst as cst
from libcst.metadata import ParentNodeProvider, ScopeProvider
from parameterized import parameterized

from opine.metadata import Distribution, SetupCallAnalyzer, from_setup_py, main


class DistributionTest(unittest.TestCase):
    def test_added_item_in_iter(self) -> None:
        d = Distribution()
        d.setup_requires = ["a", "b"]
        self.assertIn("setup_requires", list(d))


class MetadataTest(unittest.TestCase):
    def _read(self, data: str) -> Distribution:
        with tempfile.TemporaryDirectory() as d:
            sp = Path(d, "setup.py")
            sp.write_text(data)
            return from_setup_py(Path(d), {})

    def test_smoke(self) -> None:
        d = self._read(
            """\
from setuptools import setup
setup(
    name="foo",
    version="0.1",
)
"""
        )
        self.assertEqual("foo", d.name)
        self.assertEqual("0.1", d.version)

    def test_smoke_as_import(self) -> None:
        d = self._read(
            """\
from setuptools import setup as x
x(
    name="foo",
)
"""
        )
        self.assertEqual("foo", d.name)

    def test_smoke_unknown_fail(self) -> None:
        with self.assertRaises(SyntaxError):
            self._read(
                """\
setup(
    name="foo",
)
"""
            )

    def test_lists(self) -> None:
        d = self._read(
            """\
from setuptools import setup
setup(
    requires=["a", "b"],
)
"""
        )
        self.assertEqual(["a", "b"], d.requires)

    def test_tuples(self) -> None:
        d = self._read(
            """\
from setuptools import setup
setup(
    requires=("a", "b"),
)
"""
        )
        self.assertEqual(("a", "b"), d.requires)

    def test_bool(self) -> None:
        d = self._read(
            """\
from setuptools import setup
setup(
    include_package_data=True,
)
"""
        )
        self.assertEqual(True, d.include_package_data)

    def test_local(self) -> None:
        d = self._read(
            """\
from setuptools import setup
v = "0.1"
setup(
    version=v,
)
"""
        )
        self.assertEqual("0.1", d.version)

    def test_global(self) -> None:
        d = self._read(
            """\
from setuptools import setup
v = "0.1"
def foo():
    setup(
        version=v,
    )
"""
        )
        self.assertEqual("0.1", d.version)

    def test_dict(self) -> None:
        d = self._read(
            """\
from setuptools import setup
d = dict(version = "0.1")
def foo():
    setup(**d)
"""
        )
        self.assertEqual("0.1", d.version)


class EvaluateTest(unittest.TestCase):
    @parameterized.expand(  # type: ignore
        [
            ("final = 'x'", "x"),
            ("a='x'; final=a", "x"),
            ("a={'y': 'x'}; final=a['y']", "x"),
        ]
    )
    def test_evaluate(self, input: str, expected: Any) -> None:
        module = cst.parse_module(input)
        analyzer = SetupCallAnalyzer()

        wrapper = cst.MetadataWrapper(module)
        wrapper.visit(analyzer)

        # TODO: MetadataDependent.resolve clears self.metadata
        analyzer.metadata = {
            ScopeProvider: wrapper.resolve(ScopeProvider),
            ParentNodeProvider: wrapper.resolve(ParentNodeProvider),
        }
        # TODO: picks an arbitrary scope
        scope = next(iter(analyzer.metadata[ScopeProvider].values()))
        actual = analyzer.evaluate_in_scope(cst.Name(value="final"), scope)
        self.assertEqual(expected, actual)


class MainTest(unittest.TestCase):
    @patch("opine.metadata.print")
    def test_main(self, print_mock: Any) -> None:
        buf: List[str] = []
        print_mock.side_effect = buf.append
        with tempfile.TemporaryDirectory() as d:
            sp = Path(d, "setup.py")
            sp.write_text(
                "from setuptools import setup\nsetup(setup_requires=['abc'])\n"
            )
            sys.argv = ["main", d]
            main()
        self.assertTrue('"setup_requires": ' in buf[-1])

    @patch("opine.metadata.print")
    @patch("opine.metadata.traceback.print_exc")
    def test_main_fail(self, print_exc_mock: Any, print_mock: Any) -> None:
        buf: List[str] = []
        print_mock.side_effect = lambda x, **kwargs: buf.append(x)
        with tempfile.TemporaryDirectory() as d:
            sp = Path(d, "setup.py")
            sp.write_text("from setuptools import setup\nsetup(setup_requires\n")
            sys.argv = ["main", d]
            main()
        self.assertTrue(buf[-1].startswith("Fail:"))


if __name__ == "__main__":
    unittest.main()



from .types import (
    BoolWriter,
    ConfigField,
    DictWriter,
    ListCommaWriter,
    ListCommaWriterCompat,
    ListSemiWriter,
    SectionWriter,
    SetupCfg,
)

# Not all of these are in the metadata, but for ones that are see
# https://packaging.python.org/specifications/core-metadata/ for version added
# and what it means.
SETUP_ARGS = [
    # Metadata 1.0; This ordering is the same as _METHOD_BASENAMES in
    # distutils/dist.py which handles an older version of the metadata.
    # https://docs.python.org/3/distutils/setupscript.html#additional-meta-data
    #
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#metadata
    # does a good job telling you whether it's metadata/options in setup.cfg,
    # but doesn't really tell you what they do or what the metadata keys are or
    # what metadata version they correspond to.
    ConfigField("name", SetupCfg("metadata", "name")),  # Name
    ConfigField("version", SetupCfg("metadata", "version")),  # Version
    ConfigField("author", SetupCfg("metadata", "author")),  # Author
    ConfigField("author_email", SetupCfg("metadata", "author_email")),  # Author-email
    ConfigField("license", SetupCfg("metadata", "license")),  # License
    # TODO license_file, license_files
    ConfigField("url", SetupCfg("metadata", "url")),  # Home-page
    ConfigField("description", SetupCfg("metadata", "description")),  # Summary
    ConfigField(
        "long_description", SetupCfg("metadata", "long_description")
    ),  # Description
    # keywords
    ConfigField(
        "keywords", SetupCfg("metadata", "keywords", writer_cls=ListCommaWriterCompat)
    ),
    # platforms
    # fullname
    # contact
    # contact_email
    # Metadata 1.1, supported by distutils
    # provides
    # requires
    # obsoletes
    ConfigField(
        "classifiers",
        SetupCfg("metadata", "classifiers", writer_cls=ListSemiWriter),
        sample_value=None,
    ),
    # download_url
    # Metadata 1.1
    # supported-platform (binary only?)
    # Metadata 1.2, half-supported by distutils but not written in PKG-INFO
    ConfigField("maintainer", SetupCfg("metadata", "maintainer")),  # Maintainer
    ConfigField(
        "maintainer_email", SetupCfg("metadata", "maintainer_email")
    ),  # Maintainer-email
    # Metadata 1.2, not at all supported by distutils
    ConfigField(
        "python_requires", SetupCfg("options", "python_requires"), sample_value="<4.0"
    ),  # Requires-Python
    # requires_external
    # project_url -> dict
    ConfigField(
        "project_urls",
        SetupCfg("metadata", "project_urls", writer_cls=DictWriter),
        sample_value=None,  # {"Bugtracker": "http://example.com"},
    ),
    # requires_dist
    # provides_dist (rarely used)
    # obsoletes_dist (rarely used)
    # Metadata 2.1
    # text/plain, text/x-rst, text/markdown
    # This allows charset and variant (for markdown, GFM or CommonMark)
    ConfigField(
        "long_description_content_type",
        SetupCfg("metadata", "long_description_content_type"),
    ),  # Description-Content-Type
    # provides_extra
    # Not written to PKG-INFO
    # [options]
    ConfigField(
        "zip_safe",
        SetupCfg("options", "zip_safe", writer_cls=BoolWriter),
        sample_value=None,
    ),
    # zip_safe bool
    ConfigField(
        "setup_requires",
        SetupCfg("options", "setup_requires", writer_cls=ListSemiWriter),
        sample_value=None,
    ),
    ConfigField(
        "install_requires",
        SetupCfg("options", "install_requires", writer_cls=ListSemiWriter),
        sample_value=None,
    ),
    ConfigField(
        "tests_require",
        SetupCfg("options", "tests_require", writer_cls=ListSemiWriter),
        sample_value=None,
    ),
    ConfigField(
        "include_package_data",
        SetupCfg("options", "include_package_data", writer_cls=BoolWriter),
        sample_value=None,  # True,
    ),
    #
    ConfigField(
        "extras_require",
        SetupCfg("options.extras_require", "UNUSED", writer_cls=SectionWriter),
        sample_value=None,
    ),
    # use_2to3
    # use_2to3_fixers list-comma
    # use_2to3_exclude_fixers list-comma
    # convert_2to3_doctests list-comma
    ConfigField(
        "scripts",
        SetupCfg("options", "scripts", writer_cls=ListCommaWriter),
        sample_value=None,
    ),
    # eager_resources list-comma
    # dependency_links list-comma
    ConfigField(
        "packages",
        SetupCfg("options", "packages", writer_cls=ListCommaWriter),
        sample_value=None,
    ),
    ConfigField(
        "package_dir",
        SetupCfg("options", "package_dir", writer_cls=DictWriter),
        sample_value=None,
    ),
    ConfigField(
        "package_data",
        SetupCfg("options.package_data", "UNUSED", writer_cls=SectionWriter),
        sample_value=None,  # {"foo": ["py.typed"]},
    ),
    # package_data (section)
    # exclude_package_data (section)
    ConfigField(
        "namespace_packages",
        SetupCfg("options", "namespace_packages", writer_cls=ListCommaWriter),
        sample_value=None,  # ["foo", "bar"],
    ),
    ConfigField(
        "py_modules",
        SetupCfg("options", "py_modules", writer_cls=ListCommaWriter),
        sample_value=None,
    ),
    ConfigField(
        "data_files",
        SetupCfg("options.data_files", "UNUSED", writer_cls=SectionWriter),
        sample_value=None,
    ),
    #
    # Documented, but not in the table...
    ConfigField("test_suite", SetupCfg("options", "test_suite"), sample_value=None,),
    ConfigField("test_loader", SetupCfg("options", "test_loader"), sample_value=None,),
]



from .move_coverage_ini import MoveCoverageIni
from .move_setup_py_to_declarative import UseDeclarativeConfig
from .move_tox_ini import MoveToxIni

ALL_SUGGESTIONS = [
    UseDeclarativeConfig,
    MoveCoverageIni,
    MoveToxIni,
]



import unittest
from configparser import RawConfigParser
from io import StringIO
from typing import Dict, List, Union

from parameterized import parameterized

from imperfect import ConfigFile

from ..types import (
    BaseSuggestion,
    BoolWriter,
    DictWriter,
    ListCommaWriter,
    ListCommaWriterCompat,
    ListSemiWriter,
    SectionWriter,
    StrWriter,
)


class WriterTest(unittest.TestCase):
    @parameterized.expand(  # type: ignore
        [(False,), (True,),]
    )
    def test_bool_writer(self, arg: bool) -> None:
        c = ConfigFile()
        c.set_value("a", "b", BoolWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        self.assertEqual(str(arg).lower(), rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        [("hello",), ("a\nb\nc",),]
    )
    def test_str_writer(self, arg: str) -> None:
        c = ConfigFile()
        c.set_value("a", "b", StrWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        self.assertEqual(arg, rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        [
            ([], ""),
            (["a"], "\na"),
            (["a", "b"], "\na\nb"),
            (["a", "b", "c"], "\na\nb\nc"),
        ]
    )
    def test_list_comma_writer(self, arg: List[str], expected: str) -> None:
        c = ConfigFile()
        c.set_value("a", "b", ListCommaWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        self.assertEqual(expected, rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        [
            ([], ""),
            (["a"], "\na"),
            (["a", "b"], "\na\nb"),
            (["a", "b", "c"], "\na\nb\nc"),
        ]
    )
    def test_list_semi_writer(self, arg: List[str], expected: str) -> None:
        c = ConfigFile()
        c.set_value("a", "b", ListSemiWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        self.assertEqual(expected, rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        # fmt: off
        [
            ({}, ""),
            ({"x": "y"}, "\nx=y"),
            ({"x": "y", "z": "zz"}, "\nx=y\nz=zz"),
        ]
        # fmt: on
    )
    def test_dict_writer(self, arg: Dict[str, str], expected: str) -> None:
        c = ConfigFile()
        c.set_value("a", "b", DictWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        # I would prefer this be dangling lines
        self.assertEqual(expected, rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        # fmt: off
        [
            ([], ""),
            ("abc", "\nabc"),
            (["a"], "\na"),
            (["a", "b"], "\na\nb"),
            (["a", "b", "c"], "\na\nb\nc"),
        ]
        # fmt: on
    )
    def test_list_comma_writer_compat(
        self, arg: Union[str, List[str]], expected: str
    ) -> None:
        c = ConfigFile()
        c.set_value("a", "b", ListCommaWriterCompat().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        # I would prefer this be dangling lines
        self.assertEqual(expected, rcp["a"]["b"])

    @parameterized.expand(  # type: ignore
        [
            ([], ""),
            (["a"], "\na"),
            (["a", "b"], "\na\nb"),
            (["a", "b", "c"], "\na\nb\nc"),
        ]
    )
    def test_section_writer(self, arg: List[str], expected: str) -> None:
        c = ConfigFile()
        c.set_value("a", "b", SectionWriter().to_ini(arg))
        buf = StringIO()
        c.build(buf)

        rcp = RawConfigParser(strict=False)
        rcp.read_string(buf.getvalue())
        self.assertEqual(expected, rcp["a"]["b"])


class BaseSuggestionTest(unittest.TestCase):
    def test_base(self) -> None:
        b = BaseSuggestion()
        with self.assertRaises(NotImplementedError):
            b.check(None)  # type: ignore
        with self.assertRaisesRegex(Exception, "foo"):
            b.skip("foo")



from .move_setup_py_to_declarative import MoveSetupPyToDeclarativeTest
from .move_setup_requires import MoveSetupRequiresTest

__all__ = [
    "MoveSetupRequiresTest",
    "MoveSetupPyToDeclarativeTest",
]



from .functional import FunctionalTest
from .metadata import DistributionTest, EvaluateTest, MainTest, MetadataTest
from .setup_and_metadata import SetupArgsTest
from .types import BaseSuggestionTest, WriterTest

__all__ = [
    "FunctionalTest",
    "DistributionTest",
    "EvaluateTest",
    "MetadataTest",
    "MainTest",
    "SetupArgsTest",
    "BaseSuggestionTest",
    "WriterTest",
]



import tempfile
import unittest
from pathlib import Path
from typing import Any
from unittest.mock import patch

from ...types import Env
from ..move_setup_py_to_declarative import UseDeclarativeConfig


class MoveSetupPyToDeclarativeTest(unittest.TestCase):
    @patch("opine.suggestions.move_setup_py_to_declarative.echo_color_unified_diff")
    def test_one(self, diff: Any) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "setup.py").write_text(
                "from setuptools import setup\nsetup(name='foo')\n"
            )
            UseDeclarativeConfig().check(Env(Path(d)), True)
            # TODO unsure where the extra newline comes from
            self.assertEqual(
                "[metadata]\nname = foo\n\n", (Path(d) / "setup.cfg").read_text()
            )
            self.assertEqual(
                "from setuptools import setup\nsetup()\n",
                (Path(d) / "setup.py").read_text(),
            )



import configparser
import os
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path
from typing import Dict

from parameterized import parameterized

from ..setup_and_metadata import SETUP_ARGS
from ..types import ConfigField


def egg_info(files: Dict[str, str]) -> Dict[str, str]:
    # TODO consider
    # https://docs.python.org/3/distutils/apiref.html#distutils.core.run_setup
    # and whether that gives a Distribution that knows setuptools-only options
    with tempfile.TemporaryDirectory() as d:
        for relname, contents in files.items():
            (Path(d) / relname).write_text(contents)
        try:
            # TODO this is very slow.
            subprocess.check_output(
                [sys.executable, "setup.py", "egg_info"],
                cwd=d,
                stderr=subprocess.STDOUT,
                encoding="utf-8",
            )
        except subprocess.CalledProcessError as e:  # pragma: no cover
            raise Exception(f"Failed to run egg_info: {e.stdout}")

        sources = list(Path(d).rglob("PKG-INFO"))
        assert len(sources) == 1

        c1 = configparser.RawConfigParser(strict=False)
        # TODO figure out what the right api for this is; looks like rfc822
        c1.read_string("[config]\n" + sources[0].read_text())

    return dict(c1["config"])


# These tests do not increase coverage, and just verify that we have the right
# static data.
@unittest.skipIf("SKIP_ARGS_TEST" in os.environ, "SKIP_ARGS_TEST")
class SetupArgsTest(unittest.TestCase):
    @parameterized.expand(  # type: ignore
        [(t,) for t in SETUP_ARGS if t.sample_value is not None]
    )
    def test_arg_mapping(self, field: ConfigField) -> None:
        # Tests that the same arg from setup.py or setup.cfg makes it into
        # metadata in the same way.
        foo = field.sample_value
        setup_py = egg_info(
            {
                "setup.py": "from setuptools import setup\n"
                f"setup({field.keyword}={foo!r})\n"
            }
        )

        setup_cfg = egg_info(
            {
                "setup.cfg": f"[{field.cfg.section}]\n" f"{field.cfg.key} = {foo}\n",
                "setup.py": "from setuptools import setup\n" "setup()\n",
            }
        )

        setup_py_matching_keys = [k for k, v in setup_py.items() if v == str(foo)]
        setup_cfg_matching_keys = [k for k, v in setup_cfg.items() if v == str(foo)]

        if (
            len(setup_py_matching_keys) == 0 or len(setup_cfg_matching_keys) == 0
        ):  # pragma: no cover
            print(setup_py, setup_cfg)

        self.assertEqual(1, len(setup_py_matching_keys), setup_py_matching_keys)
        self.assertEqual(1, len(setup_cfg_matching_keys), setup_cfg_matching_keys)
